from yta_audio.parser import AudioParser
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileExtension
from pydub import AudioSegment, silence
from typing import Union


class AudioSilence:
    """
    Class to simplify and encapsulate the interaction with
    audio silences.
    """

    @staticmethod
    def detect(
        audio,
        minimum_silence_ms: int = 250
    ):
        """
        Detect the silences of a minimum of 'minimum_silence_ms'
        milliseconds time and returns an array containing tuples
        with the start and the end of the silence moments.

        This method returns an array of tuples with the start 
        and the end of each silence expressed in seconds.
        """
        audio = AudioParser.as_audiosegment(audio)

        minimum_silence_ms = 250 if not minimum_silence_ms else minimum_silence_ms

        if not NumberValidator.is_positive_number(minimum_silence_ms):
            raise Exception('The provided "minimum_silence_ms" is not a positive number.')

        dBFS = audio.dBFS
        # TODO: Why '- 16' (?) I don't know
        silences = silence.detect_silence(audio, min_silence_len = minimum_silence_ms, silence_thresh = dBFS - 16)

        # [(1.531, 1.946), (..., ...), ...] in seconds
        return [
            ((start / 1000), (stop / 1000))
            for start, stop in silences
        ]
    
    @staticmethod
    def create(
        duration: float,
        frame_rate: int = 11025,
        output_filename: Union[str, None] = None
    ):
        """
        Create a silence audio of the given 'duration'. The
        frame rate could be necessary due to different
        videos frame rates.
        
        The file will be stored locally only if
        'output_filename' parameter is provided.
        """
        if not NumberValidator.is_positive_number(duration, do_include_zero = False):
            raise Exception('The provided "duration" is not a positive number.')
        
        # This is the default value for AudioSegment
        frame_rate = 11025 if frame_rate is None else frame_rate

        silence = AudioSegment.silent(duration * 1000, frame_rate)

        """
        if 'output_filename' is True => '.Temp.create_filename()'
        if 'output_filename' is str => validate extension and/or fix it with '.Temp.create_filename()'
        if 'output_filename' is anything else => NOT STORED
        """

        if output_filename is not None:
            # TODO: Validate output and extension
            silence.export(Output.get_filename(output_filename, FileExtension.MP3), format = 'mp3')

        # TODO: Maybe return a FileReturn (?)
        return AudioParser.as_audioclip(silence)
    
__all__ = [
    'AudioSilence'
]


"""
Audio transcription made simple by using the classes
contained in this module.
"""
from yta_audio.voice.transcription.whisper import transcribe_whisper_without_timestamps, transcribe_whisper_with_timestamps, WhisperModel
from yta_audio.voice.transcription.objects import AudioTranscriptionWord, AudioTranscription, AudioTranscriptionWordTimestamp
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.file.writer import FileWriter
from yta_general_utils.temp import Temp
from abc import ABC, abstractmethod
from typing import Union, BinaryIO
from io import BytesIO

import numpy as np


class _AudioTranscriptor(ABC):
    """
    Abstract class to be inherited by audio
    transcriptors that do not include timestamps.
    """

    @staticmethod
    @abstractmethod
    def transcribe(
        audio: any,
        initial_prompt: str
    ):
        """
        Transcribe the provided 'audio' with the help of
        the 'initial_prompt' if provided and get the
        transcripted text.
        """
        pass

class _TimestampedAudioTranscriptor(ABC):
    """
    Abstract class to be inherited by audio
    transciptors that include timestamps.
    """

    @staticmethod
    @abstractmethod
    def transcribe(
        audio: any,
        initial_prompt: str
    ):
        """
        Transcribe the provided 'audio' with the help of
        the 'initial_prompt' if provided and get the
        transcripted text with the time moments in which
        each word is detected.
        """
        pass

class DefaultAudioTranscriptor(_AudioTranscriptor):
    """
    Class to make the transcription more simple by
    choosing a transcriptor for you. You don't
    know which transcriptor you want to use? Use
    this one.
    """

    @staticmethod
    def transcribe(
        audio: Union[str, BinaryIO, BytesIO, np.ndarray],
        initial_prompt: Union[str, None] = None,
    ):
        return WhisperAudioTranscriptor.transcribe(audio, initial_prompt)
    
class DefaultTimestampedAudioTranscriptor(_TimestampedAudioTranscriptor):
    """
    Class to make the timestamped transcription more
    simple by choosing a timestamped transcriptor for
    you. You don't know which transcriptor you want
    to use? Use this one.
    """

    @staticmethod
    def transcribe(
        audio: Union[str, BinaryIO, BytesIO, np.ndarray],
        initial_prompt: Union[str, None] = None,
    ):
        return WhisperTimestampedAudioTranscriptor.transcribe(audio, initial_prompt)

class WhisperAudioTranscriptor(_AudioTranscriptor):
    """
    Whisper simple audio transcriptor that does
    not give timestamps of the the words said.
    """

    @staticmethod
    def transcribe(
        audio: Union[str, BinaryIO, BytesIO, np.ndarray],
        initial_prompt: Union[str, None] = None,
        model: WhisperModel = WhisperModel.BASE
    ):
        if (
            not PythonValidator.is_string(audio) and
            not PythonValidator.is_instance(audio, [BinaryIO, BytesIO]) and
            not PythonValidator.is_numpy_array(audio)
        ):
            raise Exception('The provided "audio" parameter is not a string, a BinaryIO nor a numpy array.')
        
        if (
            initial_prompt is not None and
            not PythonValidator.is_string(initial_prompt)
        ):
            raise Exception('The parameter "initial_prompt" given is not None nor a valid string.')

        # TODO: Try to do this without writing a file because
        # we could have no permission in the remote server
        # This below is very interesting for that purpose:
        # https://community.openai.com/t/openai-whisper-send-bytes-python-instead-of-filename/84786/5
        if PythonValidator.is_instance(audio, BytesIO):
            audio = FileWriter.write_binary_file(audio.read(), Temp.create_filename('audio.wav'))
        elif PythonValidator.is_instance(audio, BinaryIO):
            audio = FileWriter.write_binary_file(audio.get_buffer(), Temp.create_filename('audio.wav'))

        transcription = transcribe_whisper_without_timestamps(audio, initial_prompt, model)

        return AudioTranscription([
            # TODO: Do I have confidence here (?)
            AudioTranscriptionWord(
                word = word,
                timestamp = None,
                confidence = None
            )
            for word in transcription.split(' ')
        ])
    
class WhisperTimestampedAudioTranscriptor(_TimestampedAudioTranscriptor):
    """
    Whisper transcriptor that gives you the 
    timestamps of each of the transcripted words.
    """

    @staticmethod
    def transcribe(
        audio: Union[str, BinaryIO, BytesIO, np.ndarray],
        initial_prompt: Union[str, None] = None,
        model: WhisperModel = WhisperModel.BASE
    ):
        if (
            not PythonValidator.is_string(audio) and
            # TODO: If this is working, set 'BytesIO' in all of them
            not PythonValidator.is_instance(audio, [BinaryIO, BytesIO]) and
            not PythonValidator.is_numpy_array(audio)
        ):
            raise Exception('The provided "audio" parameter is not a string, a BinaryIO nor a numpy array.')
        
        if (
            initial_prompt is not None and
            not PythonValidator.is_string(initial_prompt)
        ):
            raise Exception('The parameter "initial_prompt" given is not None nor a valid string.')
        
        # TODO: Try to do this without writing a file because
        # we could have no permission in the remote server
        # This below is very interesting for that purpose:
        # https://community.openai.com/t/openai-whisper-send-bytes-python-instead-of-filename/84786/5
        if PythonValidator.is_instance(audio, BytesIO):
            audio = FileWriter.write_binary_file(audio.read(), Temp.create_filename('audio.wav'))
        elif PythonValidator.is_instance(audio, BinaryIO):
            audio = FileWriter.write_binary_file(audio.get_buffer(), Temp.create_filename('audio.wav'))
            
        words, _ = transcribe_whisper_with_timestamps(audio, initial_prompt, model)

        return AudioTranscription([
            AudioTranscriptionWord(
                word = word['text'],
                timestamp = AudioTranscriptionWordTimestamp(word['start'], word['end']),
                confidence = word['confidence']
            ) for word in words
        ])


from yta_general_utils.programming.validator import PythonValidator

import numpy as np


# TODO: Maybe I should move this class because
# it is more strictly related to moviepy audio
class AudioFrameEditor:
    """
    Class to simplify and encapsulate all the functionality
    related to audio frame edition (audio frame is a numpy).
    """

    @staticmethod
    def modify_volume(audio_frame: np.ndarray, factor: int = 100):
        return change_audio_volume(audio_frame, factor)
    
def change_audio_volume(audio_frame: np.ndarray, factor: int = 100):
    """
    Change the 'audio_frame' volume by applying the
    given 'factor'.

    Based on:
    https://github.com/Zulko/moviepy/blob/master/moviepy/audio/fx/MultiplyVolume.py
    """
    if not PythonValidator.is_numpy_array(audio_frame):
        raise Exception('The provided "audio" is not a numpy array.')
    
    number_of_channels = len(list(audio_frame[0]))
    factors_array = np.array([factor for _ in range(audio_frame.shape[0])])

    if number_of_channels == 1: return np.multiply(audio_frame, factors_array)
    if number_of_channels == 2:
        return np.multiply(
            audio_frame,
            np.array([factors_array for _ in range(number_of_channels)]).T,
        )


from yta_audio.voice.generation.voice import NarrationVoice
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.coqui import narrate as narrate_coqui, CoquiNarrationVoice, LANGUAGE_OPTIONS as COQUI_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_coqui_narrator_names_by_language, EMOTION_OPTIONS as COQUI_EMOTION_OPTIONS, SPEED_OPTIONS as COQUI_SPEED_OPTIONS, PITCH_OPTIONS as COQUI_PITCH_OPTIONS, CoquiVoiceName
from yta_audio.voice.generation.google import narrate as narrate_google, GoogleNarrationVoice, LANGUAGE_OPTIONS as GOOGLE_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_google_narrator_names_by_language, EMOTION_OPTIONS as GOOGLE_EMOTION_OPTIONS, SPEED_OPTIONS as GOOGLE_SPEED_OPTIONS, PITCH_OPTIONS as GOOGLE_PITCH_OPTIONS, GoogleTld
from yta_audio.voice.generation.microsoft import narrate as narrate_microsoft, MicrosoftNarrationVoice, LANGUAGE_OPTIONS as MICROSOFT_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_microsoft_narrator_names_by_language, EMOTION_OPTIONS as MICROSOFT_EMOTION_OPTIONS, SPEED_OPTIONS as MICROSOFT_SPEED_OPTIONS, PITCH_OPTIONS as MICROSOFT_PITCH_OPTIONS, MicrosoftVoiceName
from yta_audio.voice.generation.open_voice import narrate as narrate_open_voice, OpenVoiceNarrationVoice, LANGUAGE_OPTIONS as OPEN_VOICE_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_open_voice_narrator_names_by_language, EMOTION_OPTIONS as OPEN_VOICE_EMOTION_OPTIONS, SPEED_OPTIONS as OPEN_VOICE_SPEED_OPTIONS, PITCH_OPTIONS as OPEN_VOICE_PITCH_OPTIONS, OpenVoiceVoiceName
from yta_audio.voice.generation.tetyys import narrate_tetyys, TetyysNarrationVoice, LANGUAGE_OPTIONS as TETYYS_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_tetyys_narrator_names_by_language, EMOTION_OPTIONS as TETYYS_EMOTION_OPTIONS, SPEED_OPTIONS as TETYYS_SPEED_OPTIONS, PITCH_OPTIONS as TETYYS_PITCH_OPTIONS, TetyysVoiceName
from yta_audio.voice.generation.tiktok import narrate_tiktok, TiktokNarrationVoice, LANGUAGE_OPTIONS as TIKTOK_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_tiktok_narrator_names_by_language, EMOTION_OPTIONS as TIKTOK_EMOTION_OPTIONS, SPEED_OPTIONS as TIKTOK_SPEED_OPTIONS, PITCH_OPTIONS as TIKTOK_PITCH_OPTIONS, TiktokVoiceName
from yta_audio.voice.generation.tortoise import narrate as narrate_tortoise, TortoiseNarrationVoice, LANGUAGE_OPTIONS as TORTOISE_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_tortoise_narrator_names_by_language, EMOTION_OPTIONS as TORTOISE_EMOTION_OPTIONS, SPEED_OPTIONS as TORTOISE_SPEED_OPTIONS, PITCH_OPTIONS as TORTOISE_PITCH_OPTIONS, TortoiseVoiceName
from yta_audio.voice.generation.ttsmp3 import narrate_tts3, Ttsmp3NarrationVoice, LANGUAGE_OPTIONS as TTSMP3_LANGUAGE_OPTIONS, get_narrator_names_by_language as get_ttsmp3_narrator_names_by_language, EMOTION_OPTIONS as TTSMP3_EMOTION_OPTIONS, SPEED_OPTIONS as TTSMP3_SPEED_OPTIONS, PITCH_OPTIONS as TTSMP3_PITCH_OPTIONS, Ttsmp3VoiceName
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from abc import ABC, abstractmethod
from typing import Union


class VoiceNarrator(ABC):
    """
    Class to simplify and encapsulate the voice
    narration functionality.
    """

    @staticmethod
    @abstractmethod
    def narrate(
        text: str,
        voice: NarrationVoice = NarrationVoice.default(),
        output_filename: Union[str, None] = None
    ):
        """
        Create a voice narration of the given 'text' and
        stores it locally in the 'output_filename'
        provided (or in a temporary file if not provided).
        """
        pass

    @staticmethod
    @abstractmethod
    def get_available_languages() -> list[NarrationLanguage]:
        """
        Get the list of the languages that are available
        to be used in the voice narration.

        This method must be overwritten.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        """
        Get the list of the narrator names that are
        available for the given 'language' to be used
        in the voice narration.

        This method must be overwritten.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        """
        Get the list of the emotions that are 
        available to be used when narrating the
        text.

        This method must be overwritten.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        """
        Get the list of the speeds that are 
        available to be used when narrating the
        text.

        This method must be overwritten.
        """
        pass

    @staticmethod
    @abstractmethod
    def get_available_pitches() -> list[VoicePitch]:
        """
        Get the list of the pitches that are 
        available to be used when narrating the
        text.

        This method must be overwritten.
        """
        pass

class CoquiVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: CoquiVoiceName = CoquiVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = CoquiVoiceName.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = CoquiNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_coqui(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return COQUI_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_coqui_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return COQUI_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return COQUI_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return COQUI_PITCH_OPTIONS
    
class GoogleVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: GoogleTld = GoogleTld.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = GoogleTld.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = GoogleNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_google(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return GOOGLE_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_google_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return GOOGLE_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return GOOGLE_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return GOOGLE_PITCH_OPTIONS
    
class MicrosoftVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: MicrosoftVoiceName = MicrosoftVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = MicrosoftVoiceName.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = MicrosoftNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_microsoft(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return MICROSOFT_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_microsoft_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return MICROSOFT_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return MICROSOFT_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return MICROSOFT_PITCH_OPTIONS
    
class OpenVoiceVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: OpenVoiceVoiceName = OpenVoiceVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = OpenVoiceVoiceName.to_enum(name)

        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = OpenVoiceNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_open_voice(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return OPEN_VOICE_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_open_voice_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return OPEN_VOICE_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return OPEN_VOICE_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return OPEN_VOICE_PITCH_OPTIONS
    
class TetyysVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: TetyysVoiceName = TetyysVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = TetyysVoiceName.to_enum(name)

        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = TetyysNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_tetyys(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return TETYYS_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_tetyys_narrator_names_by_language(language)

    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return TETYYS_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return TETYYS_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return TETYYS_PITCH_OPTIONS
    
class TiktokVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: TiktokVoiceName = TiktokVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = TiktokVoiceName.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = TiktokNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_tiktok(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return TIKTOK_LANGUAGE_OPTIONS

    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_tiktok_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return TIKTOK_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return TIKTOK_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return TIKTOK_PITCH_OPTIONS
    
class TortoiseVoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: TortoiseVoiceName = TortoiseVoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = TortoiseVoiceName.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = TortoiseNarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_tortoise(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return TORTOISE_LANGUAGE_OPTIONS
    
    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_tortoise_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return TORTOISE_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return TORTOISE_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return TORTOISE_PITCH_OPTIONS
    
class Ttsmp3VoiceNarrator(VoiceNarrator):

    @staticmethod
    def narrate(
        text: str,
        name: Ttsmp3VoiceName = Ttsmp3VoiceName.DEFAULT,
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ):
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        name = Ttsmp3VoiceName.to_enum(name)
        
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

        voice = Ttsmp3NarrationVoice(
            name = name.value,
            emotion = emotion,
            speed = speed,
            pitch = pitch,
            language = language
        )
        
        # TODO: Maybe return a FileReturn (?)
        return narrate_tts3(text, voice, output_filename = output_filename)
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return TTSMP3_LANGUAGE_OPTIONS

    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        language = NarrationLanguage.to_enum(language)

        return get_ttsmp3_narrator_names_by_language(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return TTSMP3_EMOTION_OPTIONS
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return TTSMP3_SPEED_OPTIONS
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return TTSMP3_PITCH_OPTIONS
    
class DefaultVoiceNarrator(VoiceNarrator):
    """
    The voice narrator that would be used by
    default when no specific narrator is requested.
    """

    @staticmethod
    def narrate(
        text: str,
        output_filename: Union[str, None] = None
    ):
        return GoogleVoiceNarrator.narrate(
            text = text,
            output_filename = output_filename
        )
    
    @staticmethod
    def get_available_languages() -> list[NarrationLanguage]:
        return GoogleVoiceNarrator.get_available_languages()

    @staticmethod
    def get_available_narrator_names(
        language: NarrationLanguage
    ) -> list[str]:
        return GoogleVoiceNarrator.get_available_narrator_names(language)
    
    @staticmethod
    def get_available_emotions() -> list[VoiceEmotion]:
        return GoogleVoiceNarrator.get_available_emotions()
    
    @staticmethod
    def get_available_speeds() -> list[VoiceSpeed]:
        return GoogleVoiceNarrator.get_available_speeds()
    
    @staticmethod
    def get_available_pitches() -> list[VoicePitch]:
        return GoogleVoiceNarrator.get_available_pitches()
    


from yta_audio.voice.enums import NarrationLanguage, VoiceSpeed, VoiceEmotion, VoicePitch
from dataclasses import dataclass
from abc import abstractmethod


@dataclass
class NarrationVoice:
    """
    Dataclass to be implemented by other custom
    dataclasses that will determine the narration
    voice parameters of our voice narration 
    engines.
    """

    name: str
    """
    The voice narration name.
    """
    emotion: VoiceEmotion
    """
    The voice narration emotion.
    """
    speed: VoiceSpeed
    """
    The voice narration desired speed.
    """
    pitch : VoicePitch
    """
    The voice narration desired pitch.
    """
    language: NarrationLanguage
    """
    The language to be used with the voice narration.
    """
    # TODO: Maybe add something more like
    # pitch or something

    def __init__(
        self,
        name: str = '',
        emotion: VoiceEmotion = VoiceEmotion.DEFAULT,
        speed: VoiceSpeed = VoiceSpeed.DEFAULT,
        pitch: VoicePitch = VoicePitch.DEFAULT,
        language: NarrationLanguage = NarrationLanguage.DEFAULT
    ):
        self.validate(name, emotion, speed, pitch, language)

        # TODO: Maybe we could receive an Enum name 
        # and we need to parse it
        self.name = name
        self.emotion = VoiceEmotion.to_enum(emotion)
        self.speed = VoiceSpeed.to_enum(speed)
        self.pitch = VoicePitch.to_enum(pitch)
        self.language = NarrationLanguage.to_enum(language)

    @abstractmethod
    def validate(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        """
        Check if the parameters provided are valid or not
        and raise an Exception if not.

        This method can also process the attributes to make
        some modifications and return them to be stored
        once they have been modified.

        This method must be overwritten.
        """
        pass

    @staticmethod
    @abstractmethod
    def default():
        """
        Return an instance of your Narration Voice custom
        class with the default values for that type of 
        class.

        This method must be overwritten.
        """
        pass




"""
This voice engine is based in Coqui and I
have another voice engine which is Coqui,
but using other voice narrators.

TODO: Consider mixing both voice engines
and appending this voice narrator to the
Coqui system and keep only one of them.

-- Update 19/04/2025 --
I've found that they created a fork in
https://github.com/idiap/coqui-ai-TTS with
a new version that is maintained, and the 
'tts' was generating conflicts.
"""
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union
from TTS.api import TTS


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class TortoiseVoiceName(Enum):
    """
    Available voices. The value is what is used
    for the audio creation.
    """

    DEFAULT = DEFAULT_VOICE
    SPANISH = 'es_002'
    MEXICAN = 'es_mx_002'
    # TODO: There a a lot of English US and more languages voices

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.SPANISH,
    NarrationLanguage.DEFAULT
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]

class TortoiseNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Tortoise engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        # TODO: We are not using voice names here
        return None

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> int:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None

    @property
    def processed_pitch(
        self
    ) -> int:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        language = (
            NarrationLanguage.SPANISH
            if self.language == NarrationLanguage.DEFAULT else
            self.language
        )

        return {
            NarrationLanguage.SPANISH: 'es'
        }[language]

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        TortoiseVoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return TortoiseNarrationVoice(
            name = TortoiseVoiceName.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )
    
# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            TortoiseVoiceName.DEFAULT.value,
        ]
    }[language]

# All the remaining functionality we need to make it
# work properly
def narrate(
    text: str,
    voice: TortoiseNarrationVoice = TortoiseNarrationVoice.default(),
    output_filename: Union[str, None] = None
):
    """
    @deprecated

    TODO: Remove this file and method if useless. Please, read below to check.
    This method should be removed and also the file as it is only one specific
    model in TTS narration library. It is not a different system. So please,
    remove it if it won't be used.
    """
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

    # TODO: Delete tortoise lib?
    # TODO: Delete en/multi-datase/tortoise-v2 model
    tts = TTS("tts_models/es/multi-dataset/tortoise-v2")

    # Check code here: https://docs.coqui.ai/en/latest/models/tortoise.html
    tts.tts_to_file(text = text, language = voice.processed_language, file_path = output_filename)

    return output_filename

    #reference_clips = [utils.audio.load_audio(p, 22050) for p in clips_paths]
    
    #pcm_audio = tts.tts(text)
    #pcm_audio = tts.tts_with_preset("your text here", voice_samples=reference_clips, preset='fast')
    
    #from tortoise.utils.audio import load_audio, load_voice


"""
This was extracted from here:
- https://gesserit.co/tiktok

And you have more projects here:
- Project to use Tiktok API and cookie (https://github.com/Steve0929/tiktok-tts)
- Pproject to use Tiktok API and session id (https://github.com/oscie57/tiktok-voice)
- Project that is install and play (I think) https://github.com/Giooorgiooo/TikTok-Voice-TTS/blob/main/tiktokvoice.py
"""
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_general_utils.text.transformer import remove_non_ascii_characters
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.file.writer import FileWriter
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union

import requests
import base64


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class TiktokVoiceName(Enum):
    """
    Available voices. The value is what is used
    for the audio creation.
    """

    DEFAULT = DEFAULT_VOICE
    SPANISH = 'es_002'
    MEXICAN = 'es_mx_002'
    # TODO: There a a lot of English US and more languages voices

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.SPANISH,
    NarrationLanguage.DEFAULT
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]

class TiktokNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Tiktok engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        # TODO: Maybe this DEFAULT value has to exist
        # for each language so it chooses one voice name
        # for that language
        return (
            TiktokVoiceName.SPANISH.value
            if TiktokVoiceName.to_enum(self.name) == TiktokVoiceName.DEFAULT else
            TiktokVoiceName.to_enum(self.name).value
        )

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> int:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None

    @property
    def processed_pitch(
        self
    ) -> int:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        # TODO: There is not language associated with this
        # narration voice engine. The language is set in
        # the voice name
        return None

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        TiktokVoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return TiktokNarrationVoice(
            name = TiktokVoiceName.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )
    
# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            TiktokVoiceName.DEFAULT.value,
            TiktokVoiceName.SPANISH.value,
            TiktokVoiceName.MEXICAN.value
        ]
    }[language]
    
# All the remaining functionality we need to make it
# work properly
def narrate_tiktok(
    text: str,
    voice: TiktokNarrationVoice = TiktokNarrationVoice.default(),
    output_filename: Union[str, None] = None
):
    """
    This is the tiktok voice based on a platform that generates it.
    This will make a narration with the tiktok voice. You can
    change the code to use the mexican voice.

    As this is based on an external platform, it could fail.
    """
    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'content-type': 'text/plain;charset=UTF-8',
        'origin': 'https://gesserit.co',
        'referer': 'https://gesserit.co/tiktok',
        'sec-ch-ua': '"Google Chrome";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
    }

    # Non-English characters are not accepted by Tiktok TTS generation, so:
    text = remove_non_ascii_characters(text)
    
    #data = f'{"text":"{text}","voice":"{voice.name}"}'
    data = '{"text":"' + text + '","voice":"' + voice.processed_name + '"}'

    base64_content = requests.post('https://gesserit.co/api/tiktok-tts', headers = headers, data = data).json()['base64']

    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
    
    return FileWriter.write_binary_file(
        binary_data = base64.b64decode(base64_content),
        output_filename = output_filename
    )
    


"""
This link could be interesting:
- https://github.com/DarinRowe/googletrans

You have a lot of information here:
- https://en.wikipedia.org/wiki/IETF_language_tag
- https://pypi.org/project/langcodes/
- https://gtts.readthedocs.io/en/latest/module.html#languages-gtts-lang
"""
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union
from gtts import gTTS


"""
This specific voice narration engine needs
specific values and a different parameter
handling.
"""

class GoogleNarrationLanguage(Enum):
    """
    The google narration languages accepted by their
    API
    """

    SPANISH = 'es'
    ENGLISH = 'en'

    @staticmethod
    def from_general_language(
        language: NarrationLanguage
    ) -> 'GoogleNarrationLanguage':
        """
        Turn a general 'language' instance into a Google
        narration language instance.
        """
        return {
            NarrationLanguage.DEFAULT: GoogleNarrationLanguage.SPANISH,
            NarrationLanguage.SPANISH: GoogleNarrationLanguage.SPANISH,
            NarrationLanguage.ENGLISH: GoogleNarrationLanguage.ENGLISH,
        }[NarrationLanguage.to_enum(language)]

"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.

If the engine doesn't have one specific
option (I mean, you cannot handle the 
narration speed, for example) we will allow
the user choose 'normal' value and it will
be handled just by ignoring it, but the user
will be able to choose it.
"""

# 1. The voices we accept, as Enums
class GoogleTld(Enum):

    DEFAULT = DEFAULT_VOICE
    SPANISH_SPAIN = 'es'
    SPANISH_MEXICO = 'com.mx'
    SPANISH_US = 'us'
    # TODO: How can I get the list of Tlds?I need it

    @staticmethod
    def from_google_language(
        language: GoogleNarrationLanguage
    ) -> 'GoogleTld':
        """
        Turn the Google narration 'language' into the
        corresponding Google TLD.
        """
        return {
            GoogleNarrationLanguage.SPANISH: GoogleTld.SPANISH_SPAIN,
            # TODO: Change this
            GoogleNarrationLanguage.ENGLISH:  GoogleTld.SPANISH_US,
        }[GoogleNarrationLanguage.to_enum(language)]

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.SPANISH,
    # TODO: Unavailable until I detect some valid TLDs
    #NarrationLanguage.ENGLISH, 
    NarrationLanguage.DEFAULT
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
    VoiceSpeed.SLOW,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]


class GoogleNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Google engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        # TODO: Maybe this DEFAULT value has to exist
        # for each language so it chooses one voice name
        # for that language
        return (
            GoogleTld.SPANISH_SPAIN.value
            if GoogleTld.to_enum(self.name) == GoogleTld.DEFAULT else
            GoogleTld.to_enum(self.name).value
        )

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> bool:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This value is actually saying if we are using
        # the slow mode or not
        return {
            VoiceSpeed.SLOW: True,
            VoiceSpeed.DEFAULT: False,
            VoiceSpeed.NORMAL: False
        }[self.speed]

    @property
    def processed_pitch(
        self
    ) -> float:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the pitch with
        # this voice
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        return GoogleNarrationLanguage.from_general_language(self.language).value

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        GoogleTld.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return GoogleNarrationVoice(
            name = GoogleTld.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )
        # TODO: This was in the previous version, remove when
        # confirmed that the above is working
        # return GoogleNarrationVoice('', '', 130, 1.0, NarrationLanguage.DEFAULT)
    
# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            GoogleTld.DEFAULT.value,
            GoogleTld.SPANISH_SPAIN.value,
            GoogleTld.SPANISH_MEXICO.value,
            GoogleTld.SPANISH_US.value
        ]
    }[language]


# All the remaining functionality we need to make it
# work properly
def narrate(
    text: str,
    voice: GoogleNarrationVoice = GoogleNarrationVoice.default(),
    output_filename: Union[str, None] = None
):
    """
    Creates an audio narration of the provided 'text' with the Google voice and stores it
    as 'output_filename'. This will use the provided 'language' language for the narration.
    """
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
    
    tld = GoogleTld.from_google_language(voice.processed_language).value

    gTTS(
        text = text,
        lang = voice.processed_language,
        tld = tld,
        slow = voice.processed_speed
    ).save(
        output_filename
    )

    return output_filename


"""
This module is actualy commented because it is
in quarantine. It is causing conflicts with 
'click' and 'httpx' libraries, and it is not
very important. So, by now, I'm avoiding it.
"""
# from yta_general_utils.file.filename import get_file_filename_without_extension
# from subprocess import run


# class AudioSeparator:
#     """
#     Class to simplify and encapsulate the functionality related to
#     extracting vocals, instruments and more from a single audio.
#     """
    
#     @staticmethod
#     def separate_audio_file_into_vocals_and_instruments(
#         audio_filename: str,
#         output_dirname: str
#     ):
#         """
#         This method will get 'audio_filename' audio file and will split
#         it into 2 different files stored in 'output_dirname' directory.
#         These 2 files will be 'vocals.wav' and 'accompaniment.wav' that
#         are the vocals and the instrumental part, respectively, and 
#         will be generated in a folder inside the 'output_dirname' 
#         provided that is the 'audio_filename' file name without the
#         extension, and will be returned by the function.

#         If 'example.wav' provided as 'audio_filename' and 'test' as 
#         'output_dirname', this method will create 
#         'test/example/vocals.wav' and 'test/example/accompaniment.wav'
#         audio files.
#         """
#         # This is using Spleeter, check this (https://github.com/deezer/spleeter/wiki/2.-Getting-started)
#         #
#         # This method needs a pre-trained model (that is downloaded from the repo)
#         # for each type (2stems, 4stems or 5stems). These are downloaded by default
#         # in 'pretrained_models' folder (project root) of the project in which code
#         # is being executed
#         #
#         # You can also do 'spleeter:4stems' and 'spleeter:5stems'
#         # 2stems ->   vocals.wav, accompaniment.wav
#         # 4stems ->   vocals.wav, drums.wav, bass.wav, piano.wav
#         # 5stems ->   vocals.wav, drums.wav, bass.wav, piano.wav and other.wav
#         #
#         # By now this method is only working for the '2stems' format so it will
#         # generate 'vocals.wav' and 'accompaniment.wav' in the provided 
#         # 'output_dirname'. We could personalize this in the future.
        
#         # TODO: We should accept any type of audio and, if needed, we
#         # write it to a local file to be read by the spleeter command
#         if not audio_filename:
#             return None
        
#         # TODO: Make this dynamic (by using temporary values)
#         if not output_dirname:
#             return None
        
#         command_str = 'spleeter separate -p spleeter:2stems -o ' + output_dirname + ' ' + audio_filename

#         try:
#             run(command_str)
#         except:
#             return False

#         if not output_dirname.endswith('/'):
#             output_dirname += '/'

#         generated_output_folder = output_dirname + get_file_filename_without_extension(audio_filename)
        
#         return generated_output_folder

#         # TODO: Please, make it work with python code, this below is not working
#         """
#         from spleeter.audio.adapter import AudioAdapter
#         from spleeter.separator import Separator

#         separator = Separator('spleeter:2stems')
#         # With this below we read file as numpy array (waveform parameter)
#         #from scipy.io.wavfile import read
#         #audiowave = read(audio_filename)
#         audio_adapter = AudioAdapter.default()
#         separator.separate_to_file(audio_descriptor = audio_filename, destination = output_dirname, audio_adapter = audio_adapter)
#         separator.join()
#         """


"""
This engine has been extracted from here:
- https://ttsmp3.com/

This voice engine has just a limit of
3.000 characters of input when generating
with normal voices, and 1.000 daily
characters when using AI. AI is disabled
by now as the limit makes it not
interesting for our purpose.
"""
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.downloader import Downloader
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union

import requests


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class Ttsmp3VoiceName(Enum):
    # Normal voices below:
    DEFAULT = DEFAULT_VOICE
    LUPE = 'Lupe' # US Spanish
    PENELOPE = 'Penelope' # US Spanish
    MIGUEL = 'Miguel' # US Spanish
    # TODO: There are more voices for the different
    # languages, so jus copy the names here and you
    # will be able to use them
    # AI voices below:
    # ALLOY = 'alloy' # female
    # ECHO = 'echo' # male
    # FABLE = 'fable' # male
    # ONYX = 'onyx' # male (deeper voice)
    # NOVA = 'nova' # female (soft)
    # SHIMMER = 'shimmer' # female

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.DEFAULT
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]

class Ttsmp3NarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Ttsmp3 engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        return (
            Ttsmp3VoiceName.MIGUEL.value
            if Ttsmp3VoiceName.to_enum(self.name) == Ttsmp3VoiceName.DEFAULT else
            Ttsmp3VoiceName.to_enum(self.name).value
        )

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> int:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None

    @property
    def processed_pitch(
        self
    ) -> int:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # This is not used here
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        # This engine has the language set in the voice
        # names so you should select a voice name that
        # is specific of the language you need
        return None

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        Ttsmp3VoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return Ttsmp3NarrationVoice(
            name = Ttsmp3VoiceName.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )

# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            Ttsmp3VoiceName.DEFAULT.value,
            Ttsmp3VoiceName.LUPE.value,
            Ttsmp3VoiceName.MIGUEL.value,
            Ttsmp3VoiceName.PENELOPE.value
        ]
    }[language]

# All the remaining functionality we need to make it
# work properly
# TODO: Check this because I don't know if this webpage is using the tts (coqui)
# library as the generator engine. If that, I have this engine in 'coqui.py' file
# so I don't need this (that is not stable because is based in http requests)
def narrate_tts3(
    text: str,
    voice: Ttsmp3NarrationVoice = Ttsmp3NarrationVoice.default(),
    output_filename: Union[str, None] = None
) -> str:
    """
    This makes a narration based on an external platform. You
    can change some voice configuration in code to make the
    voice different.

    Aparrently not limited. Check, because it has time breaks 
    and that stuff to enhance the narration.
    """
    # From here: https://ttsmp3.com/
    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'content-type': 'application/x-www-form-urlencoded',
        'origin': 'https://ttsmp3.com',
        'referer': 'https://ttsmp3.com/',
        'sec-ch-ua': '"Google Chrome";v="123", "Not:A-Brand";v="8", "Chromium";v="123"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
    }

    data = {
        'msg': text,
        'lang': voice.processed_name,
        'source': 'ttsmp3',
    }

    """
    There is an AI voices version but it has a
    daily limit of only 1.000 characters so it
    is not interesting, that is why I leave the
    code but commented.

    The way to request AI voice narrations is
    the same, but using the AI url and the AI
    voices names instead of the normal ones.
    """
    # AI_VERSION_HEADERS = {
    #     'accept': '*/*',
    #     'accept-language': 'es-ES,es;q=0.9',
    #     'content-type': 'application/x-www-form-urlencoded',
    #     'origin': 'https://ttsmp3.com',
    #     'priority': 'u=1, i',
    #     'referer': 'https://ttsmp3.com/ai',
    #     'sec-ch-ua': '"Not(A:Brand";v="99", "Google Chrome";v="133", "Chromium";v="133"',
    #     'sec-ch-ua-mobile': '?0',
    #     'sec-ch-ua-platform': '"Windows"',
    #     'sec-fetch-dest': 'empty',
    #     'sec-fetch-mode': 'cors',
    #     'sec-fetch-site': 'same-origin',
    #     'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36',
    # }
    # AI_VERSION_URL = 'https:ttsmp3.com/makemp3_ai.php'

    response = requests.post('https://ttsmp3.com/makemp3_new.php', headers = headers, data = data)
    response = response.json()
    url = response['URL']

    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

    # This is one example of a valid url we receive
    # as response:
    # https://ttsmp3.com/created_mp3/8b38a5f2d4664e98c9757eb6db93b914.mp3
    return Downloader.download_audio(url, output_filename).filename


from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from openvoice import se_extractor
from openvoice.api import ToneColorConverter
from melo.api import TTS
from pathlib import Path
from typing import Union

import os
import torch


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class OpenVoiceVoiceName(Enum):
    """
    Available voices. The value is what is used
    for the audio creation.
    """

    DEFAULT = DEFAULT_VOICE

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.SPANISH,
    NarrationLanguage.DEFAULT
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]

class OpenVoiceNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    OpenVoice engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        # TODO: Learn how to handle speaker ids please
        # We are not able to handle voice names until we
        # discover how the speakers ids work
        return None

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> float:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This value is used internally with numpy to
        # concatenate audios, but results may vary
        # according to the language, so this values
        # are very experimental
        speed = (
            VoiceSpeed.NORMAL
            if self.speed == VoiceSpeed.DEFAULT else
            self.speed
        )

        return {
            VoiceSpeed.SLOW: 0.8,
            VoiceSpeed.NORMAL: 1.0,
            VoiceSpeed.FAST: 1.2
        }[speed]

    @property
    def processed_pitch(
        self
    ) -> float:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the pitch with
        # this voice
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        # TODO: I don't know which values are actually
        # accepted by this voice narrator
        language = (
            NarrationLanguage.SPANISH
            if self.language == NarrationLanguage.DEFAULT else
            self.language
        )

        return {
            NarrationLanguage.SPANISH: 'ES'
        }[language]

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        OpenVoiceVoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return OpenVoiceNarrationVoice(
            name = OpenVoiceVoiceName.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )

# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            DEFAULT_VOICE,
        ]
    }[language]

# All the remaining functionality we need to make it
# work properly
def narrate(
    text: str,
    voice: OpenVoiceNarrationVoice = OpenVoiceNarrationVoice.default(),
    output_filename: Union[str, None] = None
):
    """
    Narrates the provided 'text' at the provided 'speed' with the MeloTTS
    library. The file will be saved as 'output_filename'.

    # TODO: @definitive_cantidate
    """
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
    
    model = TTS(language = voice.processed_language)
    # TODO: Find a list with the speaker IDs to
    # know how they work and how to customize it
    speaker_ids = model.hps.data.spk2id
    # TODO: What is 'quiet' for? And some other 
    # parameters? Is any interesting (?)
    model.tts_to_file(
        text = text,
        speaker_id = speaker_ids['ES'],
        output_path = output_filename,
        speed = voice.processed_speed
    )

    return output_filename

PROJECT_ABSOLUTE_PATH = get_project_abspath()

def clone_voice(input_filename):
    CHECKPOINTS_PATH = (Path(__file__).parent.parent.__str__() + '/resources/openvoice/checkpoints_v2/').replace('\\', '/')
    
    ckpt_converter = CHECKPOINTS_PATH + 'converter'
    device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')
    source_se = torch.load(f'{CHECKPOINTS_PATH}/base_speakers/ses/es.pth', map_location = device)
    # This will generate a 'se.pth' file and some wavs that are the cloned voice
    target_se, audio_name = se_extractor.get_se(input_filename, tone_color_converter, vad = False)


def imitate_voice(text, input_filename = None, output_filename = None):
    """
    This method imitates the 'input_filename' provided voice and
    generates a new narration of the provided 'text' and stores it
    as 'output_filename'.

    The provided 'input_filename' must be a valid audio file that
    contains a clear narration to be imitated.

    # TODO: @definitive_cantidate
    """
    if not input_filename:
        return None
    
    if not output_filename:
        return None
    
    CHECKPOINTS_PATH = (Path(__file__).parent.parent.__str__() + '/resources/openvoice/checkpoints_v2/').replace('\\', '/')
    
    ckpt_converter = CHECKPOINTS_PATH + 'converter'
    device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')

    source_se = torch.load(f'{CHECKPOINTS_PATH}/base_speakers/ses/es.pth', map_location = device)
    target_se, audio_name = se_extractor.get_se(input_filename, tone_color_converter, vad = False)

    # This below is for testing
    # audio_segs is the number of audio segments created
    # se_save_path is the path in which se.pth file has been saved
    # TODO: Need to know the path in which everything is saved to detect audio
    # segments number and also to be able to load the 'se.pth' file
    path = PROJECT_ABSOLUTE_PATH + 'processed/narracion_irene_albacete_recortado_v2_OMR2KXcN3jYVFUsb'
    tone_color_converter.extract_se(30, se_save_path = path), 'narracion_irene_albacete_recortado_v2_OMR2KXcN3jYVFUsb'
    # TODO: Check what is 'target_se' to check if it is a string and we can
    # point the 'se.pth' file, because I don't already understand how it works
    # This above is for testing

    # We generate a narration to obtain it but with the 'input_filename' voice
    source_filename = 'tmp.wav'
    narrate(text, output_filename = source_filename)

    encode_message = "@MyShell"
    tone_color_converter.convert(
        audio_src_path = source_filename, 
        src_se = source_se, 
        tgt_se = target_se, 
        output_path = output_filename,
        message = encode_message)
    
    # TODO: Remove tmp file 'source_filename'
    try:
        os.remove('tmp.wav')
    except:
        pass

    return output_filename



def __test():
    # TODO: This must be deleted, I keep it to ensure nothing will fail in the future
    # TODO: Took from here (https://github.com/myshell-ai/OpenVoice/blob/main/demo_part3.ipynb)
    PATH = 'C:/Users/dania/Desktop/PROYECTOS/yta-ai-utils/yta_ai_utils/'

    ckpt_converter = PATH + 'resources/openvoice/checkpoints_v2/converter'
    device = "cuda:0" if torch.cuda.is_available() else "cpu"
    output_dir = 'output/openvoice'

    tone_color_converter = ToneColorConverter(f'{ckpt_converter}/config.json', device = device)
    tone_color_converter.load_ckpt(f'{ckpt_converter}/checkpoint.pth')

    os.makedirs(output_dir, exist_ok = True)

    reference_speaker = PATH + 'resources/test.m4a' # This is the voice you want to clone
    target_se, audio_name = se_extractor.get_se(reference_speaker, tone_color_converter, vad = False)

    texts = {
        'EN_NEWEST': "Did you ever hear a folk tale about a giant turtle?",  # The newest English base speaker model
        'EN': "Did you ever hear a folk tale about a giant turtle?",
        'ES': "El resplandor del sol acaricia las olas, pintando el cielo con una paleta deslumbrante.",
        'FR': "La lueur dorée du soleil caresse les vagues, peignant le ciel d'une palette éblouissante.",
        'ZH': "在这次vacation中，我们计划去Paris欣赏埃菲尔铁塔和卢浮宫的美景。",
        'JP': "彼は毎朝ジョギングをして体を健康に保っています。",
        'KR': "안녕하세요! 오늘은 날씨가 정말 좋네요.",
    }

    src_path = f'{output_dir}/tmp.wav'

    # Basic (no cloning) below
    speed = 1.0

    for language, text in texts.items():
        model = TTS(language=language, device=device)
        speaker_ids = model.hps.data.spk2id
        
        for speaker_key in speaker_ids.keys():
            speaker_id = speaker_ids[speaker_key]
            speaker_key = speaker_key.lower().replace('_', '-')
            
            source_se = torch.load(f'{PATH}checkpoints_v2/base_speakers/ses/{speaker_key}.pth', map_location=device)
            model.tts_to_file(text, speaker_id, src_path, speed = speed)
            save_path = f'{output_dir}/output_v2_{speaker_key}.wav'

            # Run the tone color converter
            encode_message = "@MyShell"
            tone_color_converter.convert(
                audio_src_path=src_path, 
                src_se=source_se, 
                tgt_se=target_se, 
                output_path=save_path,
                message=encode_message)


from yta_audio.silences import AudioSilence
from yta_multimedia.resources.audio.drive_urls import TYPING_KEYBOARD_3_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from yta_multimedia.resources import Resource
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from moviepy import AudioFileClip, concatenate_audioclips
from typing import Union


class SoundGenerator:

    # TODO: Move this to a consts.py file
    TYPING_SOUND_FILENAME = EFFECTS_RESOURCES_FOLDER + 'sounds/typing_keyboard_3s.mp3'

    @staticmethod
    def create_typing_audio(
        output_filename: Union[str, None] = None
    ):
        """
        Creates a typing audioclip of 3.5 seconds that, if 
        'output_filename' is provided, is stored locally
        with that name.
        """
        audio_filename = Resource.get(TYPING_KEYBOARD_3_SECONDS_GOOGLE_DRIVE_DOWNLOAD_URL, cls.TYPING_SOUND_FILENAME)
        audioclip = AudioFileClip(audio_filename)
        silence_audioclip = AudioSilence.create(0.5)

        audioclip = concatenate_audioclips([audioclip, silence_audioclip])

        if output_filename is not None:
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
            audioclip.write_audiofile(output_filename)

        # TODO: Maybe use FileReturn instead (?)
        return audioclip



"""
Utils for audio conversion.
"""
from yta_general_utils.temp import Temp
from yta_general_utils.programming.validator import PythonValidator
from moviepy import AudioFileClip, AudioClip
from pydub import AudioSegment

import io
import scipy.io.wavfile as wavfile
import numpy as np


def audiosegment_to_audioclip(audio: AudioSegment):
    """
    Export the 'audio' AudiOSegment to a file and
    read it as a moviepy audio file.

    TODO: Please, make it through memory and not writting files.
    """
    if not PythonValidator.is_instance(audio, AudioSegment):
        raise Exception('The "audio" parameter provided is not an AudioSegment.')
    
    # TODO: I have not been able to create an AudioFileClip dinamically
    # from memory information. I don't want to write but...
    tmp_filename = Temp.create_filename('tmp_audio.wav')
    audio.export(tmp_filename, format = 'wav')

    return AudioFileClip(tmp_filename)

def audioclip_to_audiosegment(audio: AudioClip):
    """
    This method returns the provided moviepy audio converted into a
    pydub AudioSegment.

    TODO: This method currently writes a temporary file to make the 
    conversion. This needs to be improved to avoid writting files.
    """
    if not PythonValidator.is_instance(audio, AudioClip):
        raise Exception('The provided "audio" parameter is not an AudioClip.')

    # TODO: Please, improve this to be not writting files
    tmp_filename = Temp.create_filename('tmp_audio.wav')
    audio.write_audiofile(tmp_filename)
    audio = AudioSegment.from_file(tmp_filename, format = 'wav')

    return audio

# TODO: This has not been tested properly
def numpy_to_audiosegment(audio: np.ndarray, sample_rate):
    """
    Convers the provided 'audio' numpy array,
    that contains the audio data and must be in
    float32 or int16, to a pydub AudioSegment.

    TODO: Untested method
    """
    if not PythonValidator.is_instance(audio, np.ndarray):
        raise Exception('The "audio" parameter provided is not an np.ndarray.')
    
    # Normalize audio_array if it's not already in int16 format
    if audio.dtype != np.int16:
        if audio.dtype != np.float32:
            raise Exception('The "audio" parameter provided is not np.int16 nor np.float32.')
        
        # Assuming the audio_array is in float32 with values between -1 and 1
        audio = (audio * 32767).astype(np.int16)
    
    with io.BytesIO() as buffer:
        wavfile.write(buffer, sample_rate, audio)
        buffer.seek(0)
        
        audio_segment = AudioSegment.from_file(buffer, format = 'wav')
    
    return audio_segment

def audiosegment_to_numpy(audio: AudioSegment):
    """
    This method turns the provided 'audio' AudioSegment into a numpy
    array by converting it first to an AudioFileClip and then to a
    numpy.

    TODO: Please, maybe it is a better (more direct) way

    TODO: This method has not been tested properly
    """
    if not PythonValidator.is_instance(audio, AudioSegment):
        raise Exception('The provided "audio" parameter is not an AudioSegment.')
    
    # TODO: Maybe this is not the best way, I need
    # to test and improve this
    return np.array(audio.get_array_of_samples())

def numpy_to_audioclip(audio: np.ndarray):
    """
    This method turns the provided 'audio' np.ndarray
    into a moviepy AudioClip.

    TODO: This method has not been tested properly
    """
    if not PythonValidator.is_instance(audio, np.ndarray):
        raise Exception('The provided "audio" parameter is not an np.ndarray.')
    
    # TODO: This is untested, make it work
    return AudioClip(audio)

def audioclip_to_numpy(audio: AudioClip):
    """
    Convers the provided 'audio' moviepy AudioFileClip to a numpy
    array that will be np.float32.
    """
    if not PythonValidator.is_instance(audio, AudioClip):
        raise Exception('The provided "audio" parameter is not an AudioClip.')
    
    # TODO: Check this: https://github.com/Zulko/moviepy/issues/2027#issuecomment-1937357458

    chunk_size = 5 * 1024 * 1024
    audio_chunks = []
    for chunk in audio.iter_chunks(chunksize = chunk_size):
        # Convertir cada fragmento a un array numpy y añadirlo a la lista
        audio_array = np.frombuffer(chunk, dtype=np.int16)
        
        # Normalizar si el audio es estéreo (tendría dos columnas)
        if len(audio_array) > 0 and len(audio_array) % 2 == 0:
            audio_array = audio_array.reshape(-1, 2)
        
        audio_chunks.append(audio_array)
    
    # Concatenar todos los fragmentos en un solo array
    full_audio_array = np.concatenate(audio_chunks, axis = 0)
    
    # Convertir a float32 y normalizar
    full_audio_array = full_audio_array.astype(np.float32)
    if np.max(np.abs(full_audio_array)) > 1.0:
        full_audio_array /= np.max(np.abs(full_audio_array))
    
    return full_audio_array


def is_audio_valid(audio: list[str, np.ndarray, AudioSegment, AudioFileClip]):
    if not PythonValidator.is_string(audio) and not PythonValidator.is_instance(audio, [np.ndarray, AudioSegment, AudioFileClip]):
        return False
    
    return True

def validate_audio(audio: list[str, np.ndarray, AudioSegment, AudioFileClip]):
    # TODO: Not only 'AudioFileClip', better 'AudioClip'
    if not is_audio_valid(audio):
        raise Exception('The provided "audio" is not valid. It must be one of these clases: "str, np.ndarray, AudioSegment, AudioFileClip".')
    


__all__ = [
    'audiosegment_to_audioclip',
    'audioclip_to_audiosegment',
    'numpy_to_audiosegment',
    'audiosegment_to_numpy',
    'numpy_to_audioclip',
    'audioclip_to_numpy',
    'is_audio_valid',
    'validate_audio'
]


from yta_general_utils.constants import Language
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union


class NarrationLanguage(Enum):
    """
    The languages available for voice narrations.

    This list is based on the ISO-639 but not all
    these languages are available for narrations 
    and also each narration engine has its own
    languages available. This list is also manually
    set in other libraries, so please ensure it
    keeps updated.
    """

    DEFAULT = 'default'
    """
    This value has been created for those cases
    in which there is a default language that is
    being used in the situation we are handling.

    Using this value will provide that default
    language. For example, a Youtube video can
    be in Turkish or in English as default,
    depending on the author. Using this 'default'
    value will ensure you obtain that Youtube
    video because that default language will
    always exist.
    """
    ABKHAZIAN = Language.ABKHAZIAN.value
    AFAR = Language.AFAR.value
    AFRIKAANS = Language.AFRIKAANS.value
    AKAN = Language.AKAN.value
    ALBANIAN = Language.ALBANIAN.value
    AMHARIC = Language.AMHARIC.value
    ARABIC = Language.ARABIC.value
    ARAGONESE = Language.ARAGONESE.value
    ARMENIAN = Language.ARMENIAN.value
    ASSAMESE = Language.ASSAMESE.value
    AVARIC = Language.AVARIC.value
    AVESTAN = Language.AVESTAN.value
    AYMARA = Language.AYMARA.value
    AZERBAIJANI = Language.AZERBAIJANI.value
    BAMBARA = Language.BAMBARA.value
    BASHKIR = Language.BASHKIR.value
    BASQUE = Language.BASQUE.value
    BELARUSIAN = Language.BELARUSIAN.value
    BENGALI = Language.BENGALI.value
    BISLAMA = Language.BISLAMA.value
    BOSNIAN = Language.BOSNIAN.value
    BRETON = Language.BRETON.value
    BULGARIAN = Language.BULGARIAN.value
    BURMESE = Language.BURMESE.value
    CATALAN = Language.CATALAN.value
    CHAMORRO = Language.CHAMORRO.value
    CHECHEN = Language.CHECHEN.value
    CHICHEWA = Language.CHICHEWA.value
    CHINESE = Language.CHINESE.value
    CHINESE_TRADITIONAL = Language.CHINESE_TRADITIONAL.value
    # TODO: I think there are more complex values like
    # this above, but they are not in the list
    CHURCH_SLAVONIC = Language.CHURCH_SLAVONIC.value
    CHUVASH = Language.CHUVASH.value
    CORNISH = Language.CORNISH.value
    CORSICAN = Language.CORSICAN.value
    CREE = Language.CREE.value
    CROATIAN = Language.CROATIAN.value
    CZECH = Language.CZECH.value
    DANISH = Language.DANISH.value
    DIVEHI = Language.DIVEHI.value
    DUTCH = Language.DUTCH.value
    DZONGKHA = Language.DZONGKHA.value
    ENGLISH = Language.ENGLISH.value
    ESPERANTO = Language.ESPERANTO.value
    ESTONIAN = Language.ESTONIAN.value
    EWE = Language.EWE.value
    FAROESE = Language.FAROESE.value
    FIJIAN = Language.FIJIAN.value
    FINNISH = Language.FINNISH.value
    FRENCH = Language.FRENCH.value
    WESTERN_FRISIAN = Language.WESTERN_FRISIAN.value
    FULAH = Language.FULAH.value
    GAELIC = Language.GAELIC.value
    GALICIAN = Language.GALICIAN.value
    GANDA = Language.GANDA.value
    GEORGIAN = Language.GEORGIAN.value
    GERMAN = Language.GERMAN.value
    GREEK = Language.GREEK.value
    KALAALLISUT = Language.KALAALLISUT.value
    GUARANI = Language.GUARANI.value
    GUJARATI = Language.GUJARATI.value
    HAITIAN = Language.HAITIAN.value
    HAUSA = Language.HAUSA.value
    HEBREW = Language.HEBREW.value
    HERERO = Language.HERERO.value
    HINDI = Language.HINDI.value
    HIRI_MOTU = Language.HIRI_MOTU.value
    HUNGARIAN = Language.HUNGARIAN.value
    ICELANDIC = Language.ICELANDIC.value
    IDO = Language.IDO.value
    IGBO = Language.IGBO.value
    INDONESIAN = Language.INDONESIAN.value
    INTERLINGUA = Language.INTERLINGUA.value
    INTERLINGUE = Language.INTERLINGUE.value
    INUKTITUT = Language.INUKTITUT.value
    INUPIAQ = Language.INUPIAQ.value
    IRISH = Language.IRISH.value
    ITALIAN = Language.ITALIAN.value
    JAPANESE = Language.JAPANESE.value
    JAVANESE = Language.JAVANESE.value
    KANNADA = Language.KANNADA.value
    KANURI = Language.KANURI.value
    KASHMIRI = Language.KASHMIRI.value
    KAZAKH = Language.KAZAKH.value
    CENTRAL_KHMER = Language.CENTRAL_KHMER.value
    KIKUYU = Language.KIKUYU.value
    KINYARWANDA = Language.KINYARWANDA.value
    KYRGYZ = Language.KYRGYZ.value
    KOMI = Language.KOMI.value
    KONGO = Language.KONGO.value
    KOREAN = Language.KOREAN.value
    KUANYAMA = Language.KUANYAMA.value
    KURDISH = Language.KURDISH.value
    LAO = Language.LAO.value
    LATIN = Language.LATIN.value
    LATVIAN = Language.LATVIAN.value
    LIMBURGAN = Language.LIMBURGAN.value
    LINGALA = Language.LINGALA.value
    LITHUANIAN = Language.LITHUANIAN.value
    LUBA_KATANGA = Language.LUBA_KATANGA.value
    LUXEMBOURGISH = Language.LUXEMBOURGISH.value
    MACEDONIAN = Language.MACEDONIAN.value
    MALAGASY = Language.MALAGASY.value
    MALAY = Language.MALAY.value
    MALAYALAM = Language.MALAYALAM.value
    MALTESE = Language.MALTESE.value
    MANX = Language.MANX.value
    MAORI = Language.MAORI.value
    MARATHI = Language.MARATHI.value
    MARSHALLESE = Language.MARSHALLESE.value
    MONGOLIAN = Language.MONGOLIAN.value
    NAURU = Language.NAURU.value
    NAVAJO = Language.NAVAJO.value
    NORTH_NDEBELE = Language.NORTH_NDEBELE.value
    SOUTH_NDEBELE = Language.SOUTH_NDEBELE.value
    NDONGA = Language.NDONGA.value
    NEPALI = Language.NEPALI.value
    NORWEGIAN = Language.NORWEGIAN.value
    NORWEGIAN_BOKMAL = Language.NORWEGIAN_BOKMAL.value
    NORWEGIAN_NYNORSK = Language.NORWEGIAN_NYNORSK.value
    OCCITAN = Language.OCCITAN.value
    OJIBWA = Language.OJIBWA.value
    ORIYA = Language.ORIYA.value
    OROMO = Language.OROMO.value
    OSSETIAN = Language.OSSETIAN.value
    PALI = Language.PALI.value
    PASHTO = Language.PASHTO.value
    PERSIAN = Language.PERSIAN.value
    POLISH = Language.POLISH.value
    PORTUGUESE = Language.PORTUGUESE.value
    PUNJABI = Language.PUNJABI.value
    QUECHUA = Language.QUECHUA.value
    ROMANIAN = Language.ROMANIAN.value
    ROMANSH = Language.ROMANSH.value
    RUNDI = Language.RUNDI.value
    RUSSIAN = Language.RUSSIAN.value
    NORTHERN_SAMI = Language.NORTHERN_SAMI.value
    SAMOAN = Language.SAMOAN.value
    SANGO = Language.SANGO.value
    SANSKRIT = Language.SANSKRIT.value
    SARDINIAN = Language.SARDINIAN.value
    SERBIAN = Language.SERBIAN.value
    SHONA = Language.SHONA.value
    SINDHI = Language.SINDHI.value
    SINHALA = Language.SINHALA.value
    SLOVAK = Language.SLOVAK.value
    SLOVENIAN = Language.SLOVENIAN.value
    SOMALI = Language.SOMALI.value
    SOUTHERN_SOTHO = Language.SOUTHERN_SOTHO.value
    SPANISH = Language.SPANISH.value
    SUNDANESE = Language.SUNDANESE.value
    SWAHILI = Language.SWAHILI.value
    SWATI = Language.SWATI.value
    SWEDISH = Language.SWEDISH.value
    TAGALOG = Language.TAGALOG.value
    TAHITIAN = Language.TAHITIAN.value
    TAJIK = Language.TAJIK.value
    TAMIL = Language.TAMIL.value
    TATAR = Language.TATAR.value
    TELUGU = Language.TELUGU.value
    THAI = Language.THAI.value
    TIBETAN = Language.TIBETAN.value
    TIGRINYA = Language.TIGRINYA.value
    TONGA = Language.TONGA.value
    TSONGA = Language.TSONGA.value
    TSWANA = Language.TSWANA.value
    TURKISH = Language.TURKISH.value
    TURKMEN = Language.TURKMEN.value
    TWI = Language.TWI.value
    UIGHUR = Language.UIGHUR.value
    UKRAINIAN = Language.UKRAINIAN.value
    URDU = Language.URDU.value
    UZBEK = Language.UZBEK.value
    VENDA = Language.VENDA.value
    VIETNAMESE = Language.VIETNAMESE.value
    VOLAPUK = Language.VOLAPUK.value
    WALLOON = Language.WALLOON.value
    WELSH = Language.WELSH.value
    WOLOF = Language.WOLOF.value
    XHOSA = Language.XHOSA.value
    SICHUAN_YI = Language.SICHUAN_YI.value
    YIDDISH = Language.YIDDISH.value
    YORUBA = Language.YORUBA.value
    ZHUANG = Language.ZHUANG.value
    ZULU = Language.ZULU.value

# Engine > Language > NarratorName > Speed | Emotion
   
class VoiceEmotion(Enum):
    """
    The emotion to be transmited in the voice
    narration.
    """

    DEFAULT = 'default'
    SAD = 'sad'
    NORMAL = 'normal'
    HAPPY = 'happy'
    # TODO: Add more when available

class VoiceSpeed(Enum):
    """
    The speed to be used within the voice narration.
    """

    DEFAULT = 'default'
    SLOW = 'slow'
    NORMAL = 'normal'
    FAST = 'fast'
    # TODO: Add more when available

class VoicePitch(Enum):
    """
    The pitch to be used within the voice narration.
    """

    DEFAULT = 'default'
    LOW = 'low'
    NORMAL = 'normal'
    HIGH = 'high'
    # TODO: Add more when available

class VoiceNarrationEngine(Enum):
    """
    The engines we have available for voice narration
    generation.
    """

    DEFAULT = 'default'
    """
    When this option is provided, the system will
    choose one of the available enum elements.
    """
    COQUI = 'coqui'
    GOOGLE = 'google'
    MICROSOFT = 'microsoft'
    OPEN_VOICE = 'open_voice'
    TETYYS = 'tetyys'
    TIKTOK = 'tiktok'
    TORTOISE = 'tortoise'
    TTSMP3 = 'ttsmp3'

    def _get_engine(
        self
    ) -> 'VoiceNarrationEngine':
        """
        We turn the DEFAULT instance into a specific
        one to simplify the way we handle the options.

        For internal use only.
        """
        return (
            VoiceNarrationEngine.GOOGLE
            if self is VoiceNarrationEngine.DEFAULT else
            self
        )

    def get_voice_narrator_class(
        self
    ) -> Union['CoquiVoiceNarrator', 'GoogleVoiceNarrator', 'MicrosoftVoiceNarrator', 'OpenVoiceVoiceNarrator', 'TetyysVoiceNarrator', 'TiktokVoiceNarrator', 'TortoiseVoiceNarrator', 'Ttsmp3VoiceNarrator']:
        """
        Get the VoiceNarrator class associated with this
        enum instance.
        """
        from yta_audio.voice.generation.narrator import CoquiVoiceNarrator, GoogleVoiceNarrator, MicrosoftVoiceNarrator, OpenVoiceVoiceNarrator, TetyysVoiceNarrator, TiktokVoiceNarrator, TortoiseVoiceNarrator, Ttsmp3VoiceNarrator

        engine = self._get_engine()

        return {
            VoiceNarrationEngine.GOOGLE: GoogleVoiceNarrator,
            VoiceNarrationEngine.COQUI: CoquiVoiceNarrator,
            VoiceNarrationEngine.MICROSOFT: MicrosoftVoiceNarrator,
            VoiceNarrationEngine.OPEN_VOICE: OpenVoiceVoiceNarrator,
            VoiceNarrationEngine.TETYYS: TetyysVoiceNarrator,
            VoiceNarrationEngine.TIKTOK: TiktokVoiceNarrator,
            VoiceNarrationEngine.TORTOISE: TortoiseVoiceNarrator,
            VoiceNarrationEngine.TTSMP3: Ttsmp3VoiceNarrator,
        }[engine]
    
    @property
    def available_languages(
        self
    ) -> list[NarrationLanguage]:
        return self.get_voice_narrator_class().get_available_languages()
    
    def is_language_valid(
        self,
        language: NarrationLanguage
    ) -> bool:
        """
        Check if the given 'language' is accepted by
        this engine.
        """
        return NarrationLanguage.to_enum(language) in self.available_languages
    
    def get_available_narrator_names(
        self,
        language: NarrationLanguage
    ) -> list[str]:
        return self.get_voice_narrator_class().get_available_narrator_names(language)
    
    def is_narrator_name_valid(
        self,
        language: NarrationLanguage,
        narration_name: str
    ) -> bool:
        return narration_name in self.get_available_narrator_names(language)

    @property
    def available_emotions(
        self
    ) -> list[VoiceEmotion]:
        return self.get_voice_narrator_class().get_available_emotions()
    
    def is_emotion_valid(
        self,
        emotion: VoiceEmotion
    ) -> bool:
        return VoiceEmotion.to_enum(emotion) in self.available_emotionsWW

    @property
    def available_speeds(
        self
    ) -> list[VoiceSpeed]:
        return self.get_voice_narrator_class().get_available_speeds()
    
    def is_speed_valid(
        self,
        speed: VoiceSpeed
    ) -> bool:
        return VoiceSpeed.to_enum(speed) in self.available_speeds
    
    @property
    def available_pitches(
        self
    ) -> list[VoicePitch]:
        return self.get_voice_narrator_class().get_available_pitches()
    
    def is_pitch_valid(
        self,
        pitch: VoicePitch
    ) -> bool:
        return VoicePitch.to_enum(pitch) in self.available_pitches


from yta_audio.parser import AudioParser
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.file.filename import get_file_extension
from moviepy.Clip import Clip

from typing import Union


class AudioChannel(Enum):
    LEFT = 1
    RIGHT = 0

# TODO: Do not use yta_multimedia.video library here
# TODO: Refactor this and move to a class
def isolate_audio_channel(
    audio,
    channel: AudioChannel = AudioChannel.LEFT,
    output_filename: Union[str, None] = None
):
    """
    Gets the provided 'audio' and isolates it to he given 'channel' onlye (that can
    be left or right). It will be stored as a local file if 'output_filename' 
    provided, and will return the new isolated audio as a pydub AudioSegment.
    """
    audio = AudioParser.as_audiosegment(audio)
    channel = AudioChannel.to_enum(channel)

    channel_pan = -1.0
    if AudioChannel == AudioChannel.RIGHT:
        channel_pan = 1.0
    
    audio = adjust_audio_channels(audio, channel_pan, None, 0, audio.duration_seconds * 1000)

    if output_filename:
        output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        audio.export(out_f = output_filename, format = get_file_extension(output_filename))

    return audio

def apply_8d_effect(audio):
    """
    Generates a 8d sound effect by splitting the 'audio'' into multiple 
    smaller pieces, pans each piece to make the sound source seem like 
    it is moving from L to R and R to L in loop, decreases volume towards
    center position to make the movement sound like it is a circle 
    instead of straight line.
    """
    audio = AudioParser.as_audiosegment(audio)

    SCREEN_SIZE = MOVIEPY_SCENE_DEFAULT_SIZE[0]
    NUM_OF_PARTS = 80
    AUDIO_PART_SCREEN_SIZE = SCREEN_SIZE / NUM_OF_PARTS
    AUDIO_PART_TIME = audio.duration_seconds * 1000 / NUM_OF_PARTS

    cont = 0
    while ((cont * AUDIO_PART_TIME) < audio.duration_seconds * 1000):
        coordinate = cont * AUDIO_PART_SCREEN_SIZE
        channel_pan = x_coordinate_to_channel_pan(coordinate)
        volume_adjustment = 5 - (abs(channel_pan) / NUM_OF_PARTS) * 5

        start_time = cont * AUDIO_PART_TIME
        end_time = (cont + 1) * AUDIO_PART_TIME
        # I do this because of a small error that makes it fail
        if end_time > audio.duration_seconds * 1000:
            end_time = audio.duration_seconds * 1000
        audio = adjust_audio_channels(audio, channel_pan, volume_adjustment, start_time, end_time)
        cont += 1

    return audio

def x_coordinate_to_channel_pan(x: int):
    """
    This method calculates the corresponding channel pan value (between -1.0 and
    1.0) for the provided "x" coordinate (in an hypotetic scene of 1920x1080).
    This means that an "x" of 0 will generate a -1.0 value, and an "x" of 1919
    will generate a 1.0 value. Values out of limits (lower than 0 or greater
    than 1919) will be set as limit values (0 and 1919).

    This method has been created to be used in transition effects sounds, to be
    dynamically panned to fit the element screen position during the movement.
    """
    if not x and x != 0:
        raise Exception('No "x" provided.')
    
    if x < 0:
        x = 0

    if x > MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1:
        x = MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1

    return -1.0 + (x * 2.0 / MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1)

def adjust_audio_channels(
    audio,
    channel_pan: float = 0.0,
    volume_gain: float = 1.0,
    start_time = None,
    end_time = None
):
    """
    This method allows you to set the amount of 'audio' you want to be
    sounding on each of the 2 channels (speakers), right and left. The
    'channel_pan' parameter must be a value between -1.0, which means
    left channel, and 1.0, that means right channel. A value of 0 means
    that the sound will sound equally in left and right channel. A value
    of 0.5 means that it will sound 25% in left channel and 75% in right
    channel.

    The 'volume_gain', if 0, puts the fragment in silence. If 1, it has
    the same volume. If 2, the volume is twice.

    This method will apply the provided 'channel_pan' to the also provided
    'audio'. The 'start_time' and 'end_time' parameters determine the part
    of the audio you want the channel panning to be applied, and it is in
    seconds.
    """
    audio = AudioParser.as_audiosegment(audio)

    if not channel_pan:
        raise Exception('No "channel_pan" provided.')

    if channel_pan < -1.0 or channel_pan > 1.0:
        raise Exception('The "channel_pan" parameter must be a value between -1.0 and 1.0.')
    
    # TODO: Check that 'volume_adjustment' is a number, not only a float
    if volume_gain and not isinstance(volume_gain, float) and not isinstance(volume_gain, int):
        raise Exception('The "volume_gain" parameter must be a valid number.')
    
    if not start_time:
        start_time = 0

    if not end_time:
        end_time = audio.duration_seconds * 1000

    if start_time < 0:
        raise Exception('The "start_time" parameter cannot be lower than 0.')
    
    if start_time > audio.duration_seconds * 1000:
        raise Exception('The "start_time" cannot be greater than the actual "audio" duration.')
    
    if start_time > end_time:
        raise Exception('The "start_time" cannot be greater than the "end_time".')
    
    if end_time < 0:
        raise Exception('The "start_time" parameter cannot be lower than 0.')
    
    if end_time > audio.duration_seconds * 1000:
        raise Exception('The "end_time" cannot be greater than the actual "audio" duration.')
    
    if channel_pan < -1.0 or channel_pan > 1.0:
        raise Exception('The "channel_pan" parameter must be between -1.0 (left) and 1.0 (right)')

    # Process the part we want
    modified_part = audio[start_time: end_time]
    if volume_gain < 1:
        # We minimize the audio (x3) if lower to 1
        modified_part -= abs(modified_part.dBFS * (1 - volume_gain) * 3)
    else:
        modified_part += abs(modified_part.dBFS * (volume_gain - 1))
    modified_part = modified_part.pan(channel_pan)

    if start_time == 0 and end_time == audio.duration_seconds * 1000:
        audio = modified_part
    elif start_time == 0:
        audio = modified_part + audio[end_time: audio.duration_seconds * 1000]
    elif end_time == audio.duration_seconds * 1000:
        audio = audio[0: start_time] + modified_part
    else:
        audio = audio[0: start_time] + modified_part + audio[end_time: audio.duration_seconds * 1000]

    return audio

def get_audio_synchronized_with_video_by_position(
    audio,
    video: Union[str, Clip]
):
    """
    This method iterates over the whole provided 'video' and uses its
    position in each frame to synchronize that position with the also
    provided 'audio' that will adjust its pan according to it.

    This method returns the audio adjusted as a pydub AudioSegment.
    """
    audio = AudioParser.as_audiosegment(audio)
    video = VideoParser.to_moviepy(video)

    frames_number = int(video.fps * video.duration)
    frame_duration = video.duration / frames_number

    # I need to know the minimum x below 0 and the maximum above 1919
    minimum_x = 0
    maximum_x = MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1
    for i in range(frames_number):
        t = frame_duration * i
        # We want the center of the video to be used
        video_x = video.pos(t)[0] + video.w / 2
        if video_x < 0 and video_x < minimum_x:
            minimum_x = video_x
        if video_x > (MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1) and video_x > maximum_x:
            maximum_x = video_x

    for i in range(frames_number):
        t = frame_duration * i
        video_x = video.pos(t)[0] + video.w / 2

        # I want to make it sound always and skip our exception limits
        volume_gain = 1
        if video_x < 0:
            volume_gain -= abs(video_x / minimum_x)
            video_x = 0
        elif video_x > (MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1):
            volume_gain -= abs((video_x - (MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1)) / (maximum_x - (MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1)))
            video_x = (MOVIEPY_SCENE_DEFAULT_SIZE[0] - 1)

        audio = adjust_audio_channels(audio, x_coordinate_to_channel_pan(video_x), volume_gain, t * 1000, (t + frame_duration) * 1000)

    return audio

def synchronize_audio_pan_with_video_by_position(
    audio,
    video: Union[str, Clip]
):
    """
    This method synchronizes the provided 'video' with the also provided
    'audio' by using its position to adjust the pan.

    This method returns the provided 'video' with the new audio 
    synchronized.
    """
    # TODO: This was .to_audiofileclip() before,
    # remove this comment if working
    video = video.with_audio(AudioParser.as_audioclip(get_audio_synchronized_with_video_by_position(audio, video)))

    return video


from dataclasses import dataclass
from typing import Union

import json


@dataclass
class AudioTranscriptionWordTimestamp:
    """
    Class that holds the start and the end moment
    of a word said in a timestamped audio
    transcription.
    """

    start: any
    # TODO: Please set the 'start' timestamp type
    """
    The moment in which the word starts being said.
    """
    end: any
    # TODO: Please set the 'start' timestamp type
    """
    The moment in which the word ends being said.
    """

    @property
    def as_dict(
        self
    ):
        """
        Get the word timestamp as a dict.
        """
        return {
            'start': self.start,
            'end': self.end
        }

    @property
    def as_json(
        self
    ):
        """
        Get the word timestamp as a json.
        """
        return json.dumps(self.as_dict)

    def __init__(
        self,
        start: any,
        end: any
    ):
        # TODO: Please set the 'start' and 'end' timestamp type
        self.start = start
        self.end = end

@dataclass
class AudioTranscriptionWord:
    """
    Class that holds an audio transcription word
    and also its timestamp, that could be None if
    it is a non-timestamped audio transcription.
    """

    word: str
    """
    The word itself as a string.
    """
    timestamp: Union[AudioTranscriptionWordTimestamp, None]
    """
    The time moment in which the 'word' is said.
    """
    confidence: Union[float, None]
    """
    The confidence of this word being the correct
    word as a value between 0 and 1 (where 1 is 
    totally confident).
    """

    @property
    def start(
        self
    ):
        """
        The start time moment of this word.
        """
        return (
            self.timestamp.start
            if self.timestamp is not None else
            None
        )
    
    @property
    def end(
        self
    ):
        """
        The end time moment of this word.
        """
        return (
            self.timestamp.end
            if self.timestamp is not None else
            None
        )
    
    @property
    def as_dict(
        self
    ):
        """
        Get the word as a dict.
        """
        return {
            'word': self.word,
            'confidence': self.confidence,
            'start': self.start,
            'end': self.end
        }

    @property
    def as_json(
        self
    ):
        """
        Get the word as a json.
        """
        return json.dumps(self.as_dict)

    def __init__(
        self,
        word: str,
        timestamp: Union[AudioTranscriptionWordTimestamp, None] = None,
        confidence: Union[float, None] = None
    ):
        self.word = word
        self.timestamp = timestamp
        self.confidence = confidence

@dataclass
class AudioTranscription:
    """
    Class that holds information about an audio
    transcription, including words.
    """

    words: list[AudioTranscriptionWord]
    """
    The list of words.
    """

    @property
    def text(self):
        """
        Get the audio transcription as a single string
        text which is all the words concatenated.
        """
        return ' '.join([
            word.word
            for word in self.words
        ])
    
    @property
    def as_dict(
        self
    ):
        """
        Get the list of words as a dict.
        """
        return {
            'words': [
                word.as_dict
                for word in self.words
            ]
        }

    @property
    def as_json(
        self
    ):
        """
        Get the list of words as a json.
        """
        return json.dumps(self.as_dict)
    
    @property
    def is_timestamped(
        self
    ) -> bool:
        """
        Check if the words have their time moment or
        not. If the list of words is empty this will
        return False.
        """
        return (
            self.words[0].timestamp is not None
            if len(self.words) > 0 else
            False
        )



"""
You can see anything you need here:
- https://docs.coqui.ai/en/latest/

As this is the first voice generator engine,
I will explain some things here that are
important for all the voice narrator engines
that we are creating.

We have options, and we will have all the 
array options fulfilled with, at least, a
NORMAL and a DEFAULT options. This, even if
the voice narrator engine doesn't use those
options, will be handled. Then, when 
generating the voice narration, it will be
ignored by our system.

-- Update 19/04/2025 --
I've found that they created a fork in
https://github.com/idiap/coqui-ai-TTS with
a new version that is maintained, and the 
'tts' was generating conflicts.
"""
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union
from TTS.api import TTS


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class CoquiVoiceName(Enum):
    """
    Available voices. The value is what is used
    for the audio creation.
    """

    # tts_es_fastpitch_multispeaker.nemo
    # These below are the 2 Spanish models that exist
    DEFAULT = DEFAULT_VOICE
    SPANISH_MODEL_A = 'tts_models/es/mai/tacotron2-DDC'
    SPANISH_MODEL_B = 'tts_models/es/css10/vits'
    # TODO: There are more voices

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.DEFAULT,
    NarrationLanguage.SPANISH
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]

class CoquiNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Coqui engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        return (
            CoquiVoiceName.SPANISH_MODEL_A.value
            if CoquiVoiceName.to_enum(self.name) == CoquiVoiceName.DEFAULT else
            CoquiVoiceName.to_enum(self.name).value
        )

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> float:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the speed with
        # this voice
        return 1.0

    @property
    def processed_pitch(
        self
    ) -> float:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the pitch with
        # this voice
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        return self.language.value

    def validate(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        CoquiVoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return CoquiNarrationVoice(
            name = CoquiVoiceName.DEFAULT.value, 
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )

# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    """
    Get the voices that are available for the
    given 'language'.
    """
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            CoquiVoiceName.DEFAULT.value,
            CoquiVoiceName.SPANISH_MODEL_A.value,
            CoquiVoiceName.SPANISH_MODEL_B.value
        ]
    }[language]


# All the remaining functionality we need to make it
# work properly
def narrate(
    text: str,
    voice: CoquiNarrationVoice = CoquiNarrationVoice.default(),
    output_filename: Union[str, None] = None
) -> str:
    """
    Generates a narration audio file with the provided 'text' that
    will be stored as 'output_filename' file.

    This method uses a Spanish model so 'text' must be in Spanish.

    This method will take some time to generate the narration.
    """
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

    TTS(
        model_name = voice.processed_name
    ).tts_to_file(
        # TODO: Implement 'emotion', 'speed', etc. when known
        # how they work, the accepted values, etc. By now I'm
        # using the properties but with the default values
        text = text,
        speaker = None,
        language = None,
        emotion = voice.processed_emotion,
        speed = voice.processed_speed,
        file_path = output_filename
    )
    
    # TODO: This was in the previous version, remove when the
    # above is working.
    # tts = TTS(model_name = voice.name)
    # # There is 'language', 'emotion', 'speed'...
    # tts.tts_to_file(text = text, file_path = output_filename)

    return output_filename

def narrate_imitating_voice(
    text: str,
    input_filename: str,
    output_filename: Union[str, None] = None
):
    """
    Narrates the provided 'text' by imitating the provided 'input_filename'
    audio file (that must be a voice narrating something) and saves the 
    narration as 'output_filename'.

    The 'input_filename' could be an array of audio filenames.

    Language is set 'es' in code by default.

    This method will take time as it will recreate the voice parameters with
    which the narration will be created after that.

    ANNOTATIONS: This method is only copying the way the narration voice 
    talks, but not the own voice. This is not working as expected, as we are
    not cloning voices, we are just imitating the tone. We need another way
    to actually clone the voice as Elevenlabs do.
    """
    # TODO: This is not validating if audio file...
    if not input_filename:
        raise Exception('No "input_filename" provided.')
    
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

    tts = TTS("tts_models/multilingual/multi-dataset/xtts_v2")
    # This below will use the latest XTTS_v2 (needs to download the model)
    #tts = TTS('xtts')

    # TODO: Implement a way of identifying and storing the voices we create to
    # be able to use again them without recreating them twice.

    # input_filename can be an array of wav files
    # generate speech by cloning a voice using default settings
    tts.tts_to_file(text = text, file_path = output_filename, speaker_wav = input_filename, language = 'es')

    return output_filename


"""
# Very interesting (https://github.com/m-bain/whisperX) to get timestamps of each word
# But the one I use is this one (https://github.com/linto-ai/whisper-timestamped)
"""
from yta_general_utils.programming.enum import YTAEnum as Enum
from faster_whisper import WhisperModel as FasterWhisperModel
from typing import Union, BinaryIO
from io import BytesIO

import whisper_timestamped
import numpy as np


class WhisperModel(Enum):
    """
    The model of Whisper you want to use to detect
    the audio.

    See more:
    https://github.com/openai/whisper?tab=readme-ov-file#available-models-and-languages
    """
    
    TINY = 'tiny'
    """
    Trained with 39M parameters.
    Required VRAM: ~1GB.
    Relative speed: ~10x.
    """
    BASE = 'base'
    """
    Trained with 74M parameters.
    Required VRAM: ~1GB.
    Relative speed: ~7x.
    """
    SMALL = 'small'
    """
    Trained with 244M parameters.
    Required VRAM: ~2GB.
    Relative speed: ~4x.
    """
    MEDIUM = 'medium'
    """
    Trained with 769M parameters.
    Required VRAM: ~5GB.
    Relative speed: ~2x.
    """
    LARGE = 'large'
    """
    Trained with 1550M parameters.
    Required VRAM: ~10GB.
    Relative speed: ~1x.
    """
    TURBO = 'turbo'
    """
    Trained with 809M parameters.
    Required VRAM: ~6GB.
    Relative speed: ~8x.
    """

def transcribe_whisper_with_timestamps(
    audio: Union[str, BinaryIO, BytesIO, np.ndarray],
    initial_prompt: Union[str, None] = None,
    model: WhisperModel = WhisperModel.BASE
):
    """
    Transcribe the provided 'audio' using the specified
    'model' and obtains a list of 'words' (with the
    'start' and 'end' times) and the whole 'text' in
    the audio.

    This method returns the tuple (words, text).
    """
    model = WhisperModel.to_enum(model)

    # TODO: Is BinaryIO or BytesIO accepted here (?)
    audio = whisper_timestamped.load_audio(audio)
    model = whisper_timestamped.load_model(model.value, device = 'cpu')

    # See this: https://github.com/openai/openai-cookbook/blob/main/examples/Whisper_prompting_guide.ipynb
    # I do this 'initial_prompt' formatting due to some issues when using it 
    # as it was. I read in Cookguide to pass natural sentences like this below
    # and it seems to be working well, so here it is:
    if initial_prompt is not None:
        #initial_prompt = '""""' + initial_prompt + '""""'
        #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
        initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

    # 'vad' = True would remove silent parts to decrease hallucinations
    # 'detect_disfluences' detects corrections, repetitions, etc. so the
    # word prediction should be more accurate. Useful for natural narrations
    transcription = whisper_timestamped.transcribe(model, audio, language = "es", initial_prompt = initial_prompt)
    """
    'text', which includes the whole text
    'segments', which has the different segments
    'words', inside 'segments', which contains each 
    word and its 'text', 'start' and 'end'
    """
    words = [word for segment in transcription['segments'] for word in segment['words']]
    text = ' '.join([word['text'] for word in words])

    return words, text
    
def transcribe_whisper_without_timestamps(
    audio: Union[str, BinaryIO, BytesIO, np.ndarray],
    initial_prompt: Union[str, None] = None,
    model: WhisperModel = WhisperModel.BASE
) -> str:
    """
    Obtain the transcription with no timestamps, fast and
    ideal to get ideas from the text, summarize it, etc.

    I recommend you using the 'transcribe_with_timestamps'
    because the result is more complete.

    The result is just the transcription as a text.
    """
    # TODO: What if 'whisper_timestamped' is better?
    # We should use it
    model: FasterWhisperModel = FasterWhisperModel(WhisperModel.to_enum(model).value)

    segments, _ = model.transcribe(
        audio = audio,
        # TODO: Audio should be customizable
        language = 'es',
        beam_size = 5,
        initial_prompt = initial_prompt
    )

    text = ' '.join([
        segment.text
        for segment in segments
    ]).strip()

    return text

    model = WhisperModel.to_enum(model)

    # TODO: Is BinaryIO or BytesIO accepted here (?)
    audio = whisper_timestamped.load_audio(audio)
    model = whisper_timestamped.load_model(model.value, device = 'cpu')

    if initial_prompt is not None:
        #initial_prompt = '""""' + initial_prompt + '""""'
        #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
        initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

    # 'vad' = True would remove silent parts to decrease hallucinations
    # 'detect_disfluences' detects corrections, repetitions, etc. so the
    # word prediction should be more accurate. Useful for natural narrations
    transcription = whisper_timestamped.transcribe(model, audio, language = "es", initial_prompt = initial_prompt)

    words = [word for segment in transcription['segments'] for word in segment['words']]
    text = ' '.join([word['text'] for word in words])

    return text

    # TODO: This was previously written, maybe use
    # FastestWhisper or normal whisper (?)
    segments, _ = model.transcribe(audio, language = 'es', beam_size = 5, initial_prompt = initial_prompt)

    text = ' '.join([
        segment.text
        for segment in segments
    ]).strip()

    return text

# TODO: This is being tested, use carefully
def transcribe_whisper_without_timestamps_real_time(
    model: WhisperModel = WhisperModel.BASE
) -> str:
    """
    Detects the audio in real time speaking and
    returns the sentence that has been said.

    TODO: This work very slow, it is not worth it
    to detect commands in real time, from the
    moment you say the phrase until it is
    transcribed it takes a long time.
    """
    # Test real time speech recognition
    import os
    import numpy as np
    import speech_recognition as sr
    # TODO: Maybe try with 'faster_whisper' (?)
    import whisper
    import torch

    from datetime import datetime, timedelta
    from queue import Queue
    from time import sleep

    # Based on this: https://github.com/davabase/whisper_real_time/blob/master/transcribe_demo.py
    model = WhisperModel.to_enum(model).value
    # Use or not the English model
    #non_english = True
    # Energy level for mic to detect
    energy_threshold = 1000
    # Real time recording length (in seconds)
    record_timeout = 2
    # Empty space between recordings before
    # considering a new line (in seconds)
    phrase_timeout = 3

    # The last time a recording was retrieved
    # from the queue.
    phrase_time = None
    # Thread safe Queue for passing data from
    # the threaded recording callback
    data_queue = Queue()
    # We use SpeechRecognizer to record our audio
    # because it has a nice feature where it can
    # detect when speech ends
    recorder = sr.Recognizer()
    recorder.energy_threshold = energy_threshold
    # Definitely do this, dynamic energy compensation
    # lowers the energy threshold dramatically to a
    # point where the SpeechRecognizer never stops
    # recording
    recorder.dynamic_energy_threshold = False

    source = sr.Microphone(sample_rate = 16000)
    audio_model = whisper.load_model(model)

    transcription = ['']

    with source:
        recorder.adjust_for_ambient_noise(source)

    def record_callback(_, audio:sr.AudioData) -> None:
        """
        Threaded callback function to receive audio data when recordings finish.
        audio: An AudioData containing the recorded bytes.
        """
        # Grab the raw bytes and push it into the
        # thread safe queue.
        data = audio.get_raw_data()
        data_queue.put(data)

    # Create a background thread that will pass us raw
    # audio bytes. We could do this manually but
    # SpeechRecognizer provides a nice helper
    recorder.listen_in_background(source, record_callback, phrase_time_limit = record_timeout)

    # Cue the user that we're ready to go.
    # print('System is listening:\n')

    while True:
        try:
            now = datetime.utcnow()
            #now = datetime.now(datetime.astimezone(utc))
            # Pull raw recorded audio from the queue
            if not data_queue.empty():
                phrase_complete = False
                # If enough time has passed between recordings,
                # consider the phrase complete
                # Clear the current working audio buffer to
                # start over with the new data
                if (
                    phrase_time
                    and now - phrase_time > timedelta(seconds=phrase_timeout)
                ):
                    phrase_complete = True
                # This is the last time we received new audio
                # data from the queue
                phrase_time = now
                
                # Combine audio data from queue
                audio_data = b''.join(data_queue.queue)
                data_queue.queue.clear()
                
                # Convert in-ram buffer to something the model
                # can use directly without needing a temp file
                # Convert data from 16 bit wide integers to
                # floating point with a width of 32 bits
                # Clamp the audio stream frequency to a PCM
                # wavelength compatible default of 32768hz max
                audio_np = np.frombuffer(
                    audio_data,
                    dtype = np.int16
                ).astype(np.float32) / 32768.0

                # Read the transcription
                result = audio_model.transcribe(audio_np, fp16 = torch.cuda.is_available())
                text = result['text'].strip()

                # If we detected a pause between recordings, add
                # a new item to our transcription Otherwise edit
                # the existing one
                if phrase_complete:
                    transcription.append(text)
                    # TODO: This line is returning the text so we are
                    # able to accept one line in real time, but this
                    # code was initially prepared to handle continuos
                    # narration (remove return and check)
                    return text
                else:
                    transcription[-1] = text

                # Clear the console to reprint the updated
                # transcription
                os.system(
                    'cls'
                    if os.name == 'nt' else
                    'clear'
                )

                for line in transcription:
                    print(line)
                # Flush stdout.
                print('', end = '', flush = True)
            else:
                # Infinite loops are bad for processors, must sleep.
                sleep(0.25)
        except KeyboardInterrupt:
            break


from yta_general_utils.file.handler import FileHandler
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.temp import Temp
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from moviepy import AudioFileClip
from pydub import AudioSegment
from pydub.effects import speedup
from typing import Union


# TODO: Check and refactor this below
def crop_audio_file(
    audio_filename: str,
    duration: float,
    output_filename: Union[str, None] = None
):
    """
    Crops the 'audio_filename' provided to the requested 'duration'.

    This method returns the new audio 'output_filename' if valid, or
    False if it was not possible to crop it.
    """
    if not audio_filename:
        return None
    
    audio_clip = AudioFileClip(audio_filename)

    if audio_clip.duration < duration:
        # TODO: Exception, None, review this
        print('audio is shorter than requested duration')
        return False
    
    audio_clip = audio_clip.with_duration(duration)
    
    if output_filename is not None:
        audio_clip.write_audiofile(Output.get_filename(output_filename, FileTypeX.AUDIO))

    return audio_clip

def speedup_audio_file(
    audio_filename: str,
    new_duration: int,
    output_filename: Union[str, None] = None):
    """
    Receives an audio 'audio_filename' and makes it have the provided
    'new_duration' if shorter than the real one. It will speed up the
    audio to fit that new duration and will store the new audio file
    as 'output_filename'.
    """
    if not audio_filename:
        return None
    
    if not FileValidator.file_is_audio_file(audio_filename):
        return None
    
    audio = AudioFileClip(audio_filename)
    if audio.duration <= new_duration:
        return None
    
    # We calculate audio the speed_up factor 
    speed_factor = audio.duration / new_duration

    # We use a tmp file because input filename could be same as output 
    # TODO: Careful with extension
    tmp_audio_filename = Temp.create_filename('tmp_audio_shortened.wav')
    # TODO: What about format?
    sound = AudioSegment.from_file(audio_filename, format = 'mp3')
    final = speedup(sound, playback_speed = speed_factor)
    final.export(tmp_audio_filename, format = 'mp3')
    # TODO: This is giving 'PermissionError [WindError 5]' when input 
    # and output are the same

    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)

    FileHandler.rename_file(
        tmp_audio_filename,
        output_filename,
        True
    )

    return output_filename



from yta_audio.voice.enums import NarrationLanguage, VoiceEmotion, VoiceSpeed, VoicePitch
from yta_audio.voice.consts import DEFAULT_VOICE
from yta_audio.voice.generation.voice import NarrationVoice
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union

import pyttsx3


"""
The options below are specified even if we
don't use them later when processing the 
voice narration. This is to keep the same
structure for any voice narration and to
simplify the way we offer the options in
an API that is able to make requests.
"""

# 1. The voices we accept, as Enums
class MicrosoftVoiceName(Enum):
    """
    Available voices. The value is what is used
    for the audio creation.
    """

    DEFAULT = DEFAULT_VOICE
    SPANISH_SPAIN = 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Voices\Tokens\TTS_MS_ES-ES_HELENA_11.0'
    SPANISH_MEXICO = 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Speech\Voices\Tokens\TTS_MS_ES-MX_SABINA_11.0'
    # TODO: There are more voices

# 2. The languages we accept
LANGUAGE_OPTIONS = [
    NarrationLanguage.DEFAULT,
    NarrationLanguage.SPANISH
]

# 3. The emotions we accept
EMOTION_OPTIONS = [
    VoiceEmotion.DEFAULT,
    VoiceEmotion.NORMAL,
]

# 4. The speeds we accept
SPEED_OPTIONS = [
    VoiceSpeed.DEFAULT,
    VoiceSpeed.NORMAL,
    VoiceSpeed.SLOW,
    VoiceSpeed.FAST
]

# 5. The pitches we accept
PITCH_OPTIONS = [
    VoicePitch.DEFAULT,
    VoicePitch.NORMAL,
]


class MicrosoftNarrationVoice(NarrationVoice):
    """
    Voice instance to be used when narrating with
    Microsoft engine.
    """

    @property
    def processed_name(
        self
    ) -> str:
        """
        Get the usable name value from the one that has
        been set when instantiating the instance.
        """
        # TODO: Maybe this DEFAULT value has to exist
        # for each language so it chooses one voice name
        # for that language
        return (
            MicrosoftVoiceName.SPANISH_SPAIN.value
            if MicrosoftVoiceName.to_enum(self.name) == MicrosoftVoiceName.DEFAULT else
            MicrosoftVoiceName.to_enum(self.name).value
        )

    @property
    def processed_emotion(
        self
    ) -> str:
        """
        Get the usable emotion value from the one that
        has been set when instantiating the instance.
        """
        # This narration is not able to handle any 
        # emotion (at least by now)
        return None
    
    @property
    def processed_speed(
        self
    ) -> int:
        """
        Get the usable speed value from the one that
        has been set when instantiating the instance.
        """
        # This value is actually the amount of words per
        # minute to be said during the speech
        speed = (
            VoiceSpeed.NORMAL
            if self.speed == VoiceSpeed.DEFAULT else
            self.speed
        )

        return {
            VoiceSpeed.SLOW: 160,
            VoiceSpeed.NORMAL: 200,
            VoiceSpeed.FAST: 240
        }[speed]

    @property
    def processed_pitch(
        self
    ) -> float:
        """
        Get the usable pitch value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the pitch with
        # this voice
        return None
    
    @property
    def processed_language(
        self
    ) -> str:
        """
        Get the usable language value from the one that
        has been set when instantiating the instance.
        """
        # By now we are not handling the language with
        # this voice
        return None

    def validate_and_process(
        self,
        name: str,
        emotion: VoiceEmotion,
        speed: VoiceSpeed,
        pitch: VoicePitch,
        language: NarrationLanguage
    ):
        MicrosoftVoiceName.to_enum(name)
        if VoiceEmotion.to_enum(emotion) not in EMOTION_OPTIONS:
            raise Exception(f'The provided {emotion} is not valid for this narration voice.')
        if VoiceSpeed.to_enum(speed) not in SPEED_OPTIONS:
            raise Exception(f'The provided {speed} is not valid for this narration voice.')
        if VoicePitch.to_enum(pitch) not in PITCH_OPTIONS:
            raise Exception(f'The provided {pitch} is not valid for this narration voice.')
        if NarrationLanguage.to_enum(language) not in LANGUAGE_OPTIONS:
            raise Exception(f'The provided {language} is not valid for this narration voice.')
        
    @staticmethod
    def default():
        return MicrosoftNarrationVoice(
            name = MicrosoftVoiceName.DEFAULT.value,
            emotion = VoiceEmotion.DEFAULT,
            speed = VoiceSpeed.DEFAULT,
            pitch = VoicePitch.DEFAULT,
            language = NarrationLanguage.DEFAULT
        )

# The voices but for a specific language, to be able to
# choose one when this is requested from the outside
def get_narrator_names_by_language(
    language: NarrationLanguage
) -> list[str]:
    language = NarrationLanguage.to_enum(language)
    language = (
        NarrationLanguage.SPANISH
        if language is NarrationLanguage.DEFAULT else
        language
    )

    return {
        NarrationLanguage.SPANISH: [
            MicrosoftVoiceName.DEFAULT.value,
            MicrosoftVoiceName.SPANISH_SPAIN.value,
            MicrosoftVoiceName.SPANISH_MEXICO.value
        ]
    }[language]

# All the remaining functionality we need to make it
# work properly
def narrate(
    text: str,
    voice: MicrosoftNarrationVoice = MicrosoftNarrationVoice.default(),
    output_filename: Union[str, None] = None
):
    """
    Creates an audio narration of the provided 'text'
    and stores it as 'output_filename'.
    """
    output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
    
    engine = pyttsx3.init()
    engine.setProperty('voice', voice.processed_name)
    # Default speed is 200 (wpm)
    engine.setProperty('rate', voice.processed_speed)
    engine.save_to_file(text, output_filename)
    engine.runAndWait()

    return output_filename